# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/hierarchicalforecast/',
                'doc_host': 'https://Nixtla.github.io',
                'git_url': 'https://github.com/Nixtla/hierarchicalforecast/',
                'lib_path': 'hierarchicalforecast'},
  'syms': { 'hierarchicalforecast.core': { 'hierarchicalforecast.core.HierarchicalReconciliation': ( 'src/core.html#hierarchicalreconciliation',
                                                                                                     'hierarchicalforecast/core.py'),
                                           'hierarchicalforecast.core.HierarchicalReconciliation.__init__': ( 'src/core.html#hierarchicalreconciliation.__init__',
                                                                                                              'hierarchicalforecast/core.py'),
                                           'hierarchicalforecast.core.HierarchicalReconciliation._prepare_Y': ( 'src/core.html#hierarchicalreconciliation._prepare_y',
                                                                                                                'hierarchicalforecast/core.py'),
                                           'hierarchicalforecast.core.HierarchicalReconciliation._prepare_fit': ( 'src/core.html#hierarchicalreconciliation._prepare_fit',
                                                                                                                  'hierarchicalforecast/core.py'),
                                           'hierarchicalforecast.core.HierarchicalReconciliation.bootstrap_reconcile': ( 'src/core.html#hierarchicalreconciliation.bootstrap_reconcile',
                                                                                                                         'hierarchicalforecast/core.py'),
                                           'hierarchicalforecast.core.HierarchicalReconciliation.reconcile': ( 'src/core.html#hierarchicalreconciliation.reconcile',
                                                                                                               'hierarchicalforecast/core.py'),
                                           'hierarchicalforecast.core._build_fn_name': ( 'src/core.html#_build_fn_name',
                                                                                         'hierarchicalforecast/core.py'),
                                           'hierarchicalforecast.core._reverse_engineer_sigmah': ( 'src/core.html#_reverse_engineer_sigmah',
                                                                                                   'hierarchicalforecast/core.py')},
            'hierarchicalforecast.evaluation': { 'hierarchicalforecast.evaluation.HierarchicalEvaluation': ( 'src/evaluation.html#hierarchicalevaluation',
                                                                                                             'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.HierarchicalEvaluation.__init__': ( 'src/evaluation.html#hierarchicalevaluation.__init__',
                                                                                                                      'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.HierarchicalEvaluation.evaluate': ( 'src/evaluation.html#hierarchicalevaluation.evaluate',
                                                                                                                      'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation._loss_deprecation_notice': ( 'src/evaluation.html#_loss_deprecation_notice',
                                                                                                               'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation._metric_protections': ( 'src/evaluation.html#_metric_protections',
                                                                                                          'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.energy_score': ( 'src/evaluation.html#energy_score',
                                                                                                   'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.evaluate': ( 'src/evaluation.html#evaluate',
                                                                                               'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.log_score': ( 'src/evaluation.html#log_score',
                                                                                                'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.mqloss': ( 'src/evaluation.html#mqloss',
                                                                                             'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.mse': ( 'src/evaluation.html#mse',
                                                                                          'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.msse': ( 'src/evaluation.html#msse',
                                                                                           'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.rel_mse': ( 'src/evaluation.html#rel_mse',
                                                                                              'hierarchicalforecast/evaluation.py'),
                                                 'hierarchicalforecast.evaluation.scaled_crps': ( 'src/evaluation.html#scaled_crps',
                                                                                                  'hierarchicalforecast/evaluation.py')},
            'hierarchicalforecast.methods': { 'hierarchicalforecast.methods.BottomUp': ( 'src/methods.html#bottomup',
                                                                                         'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.BottomUp._get_PW_matrices': ( 'src/methods.html#bottomup._get_pw_matrices',
                                                                                                          'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.BottomUp.fit': ( 'src/methods.html#bottomup.fit',
                                                                                             'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.BottomUp.fit_predict': ( 'src/methods.html#bottomup.fit_predict',
                                                                                                     'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.BottomUpSparse': ( 'src/methods.html#bottomupsparse',
                                                                                               'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.BottomUpSparse._get_PW_matrices': ( 'src/methods.html#bottomupsparse._get_pw_matrices',
                                                                                                                'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.ERM': ( 'src/methods.html#erm',
                                                                                    'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.ERM.__init__': ( 'src/methods.html#erm.__init__',
                                                                                             'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.ERM._get_PW_matrices': ( 'src/methods.html#erm._get_pw_matrices',
                                                                                                     'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.ERM.fit': ( 'src/methods.html#erm.fit',
                                                                                        'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.ERM.fit_predict': ( 'src/methods.html#erm.fit_predict',
                                                                                                'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.HReconciler': ( 'src/methods.html#hreconciler',
                                                                                            'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.HReconciler._get_sampler': ( 'src/methods.html#hreconciler._get_sampler',
                                                                                                         'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.HReconciler._reconcile': ( 'src/methods.html#hreconciler._reconcile',
                                                                                                       'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.HReconciler.fit': ( 'src/methods.html#hreconciler.fit',
                                                                                                'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.HReconciler.fit_predict': ( 'src/methods.html#hreconciler.fit_predict',
                                                                                                        'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.HReconciler.predict': ( 'src/methods.html#hreconciler.predict',
                                                                                                    'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.HReconciler.sample': ( 'src/methods.html#hreconciler.sample',
                                                                                                   'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOut': ( 'src/methods.html#middleout',
                                                                                          'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOut.__init__': ( 'src/methods.html#middleout.__init__',
                                                                                                   'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOut._get_PW_matrices': ( 'src/methods.html#middleout._get_pw_matrices',
                                                                                                           'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOut.fit': ( 'src/methods.html#middleout.fit',
                                                                                              'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOut.fit_predict': ( 'src/methods.html#middleout.fit_predict',
                                                                                                      'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOut.predict': ( 'src/methods.html#middleout.predict',
                                                                                                  'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOutSparse': ( 'src/methods.html#middleoutsparse',
                                                                                                'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MiddleOutSparse.fit_predict': ( 'src/methods.html#middleoutsparse.fit_predict',
                                                                                                            'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTrace': ( 'src/methods.html#mintrace',
                                                                                         'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTrace.__init__': ( 'src/methods.html#mintrace.__init__',
                                                                                                  'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTrace._get_PW_matrices': ( 'src/methods.html#mintrace._get_pw_matrices',
                                                                                                          'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTrace.fit': ( 'src/methods.html#mintrace.fit',
                                                                                             'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTrace.fit_predict': ( 'src/methods.html#mintrace.fit_predict',
                                                                                                     'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTraceSparse': ( 'src/methods.html#mintracesparse',
                                                                                               'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTraceSparse.__init__': ( 'src/methods.html#mintracesparse.__init__',
                                                                                                        'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTraceSparse._get_PW_matrices': ( 'src/methods.html#mintracesparse._get_pw_matrices',
                                                                                                                'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.MinTraceSparse.fit': ( 'src/methods.html#mintracesparse.fit',
                                                                                                   'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.OptimalCombination': ( 'src/methods.html#optimalcombination',
                                                                                                   'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.OptimalCombination.__init__': ( 'src/methods.html#optimalcombination.__init__',
                                                                                                            'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDown': ( 'src/methods.html#topdown',
                                                                                        'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDown.__init__': ( 'src/methods.html#topdown.__init__',
                                                                                                 'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDown._get_PW_matrices': ( 'src/methods.html#topdown._get_pw_matrices',
                                                                                                         'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDown.fit': ( 'src/methods.html#topdown.fit',
                                                                                            'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDown.fit_predict': ( 'src/methods.html#topdown.fit_predict',
                                                                                                    'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDownSparse': ( 'src/methods.html#topdownsparse',
                                                                                              'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDownSparse._get_PW_matrices': ( 'src/methods.html#topdownsparse._get_pw_matrices',
                                                                                                               'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods.TopDownSparse.fit_predict': ( 'src/methods.html#topdownsparse.fit_predict',
                                                                                                          'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods._get_child_nodes': ( 'src/methods.html#_get_child_nodes',
                                                                                                 'hierarchicalforecast/methods.py'),
                                              'hierarchicalforecast.methods._reconcile_fcst_proportions': ( 'src/methods.html#_reconcile_fcst_proportions',
                                                                                                            'hierarchicalforecast/methods.py')},
            'hierarchicalforecast.probabilistic_methods': { 'hierarchicalforecast.probabilistic_methods.Bootstrap': ( 'src/probabilistic_methods.html#bootstrap',
                                                                                                                      'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Bootstrap.__init__': ( 'src/probabilistic_methods.html#bootstrap.__init__',
                                                                                                                               'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Bootstrap.get_prediction_levels': ( 'src/probabilistic_methods.html#bootstrap.get_prediction_levels',
                                                                                                                                            'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Bootstrap.get_prediction_quantiles': ( 'src/probabilistic_methods.html#bootstrap.get_prediction_quantiles',
                                                                                                                                               'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Bootstrap.get_samples': ( 'src/probabilistic_methods.html#bootstrap.get_samples',
                                                                                                                                  'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Normality': ( 'src/probabilistic_methods.html#normality',
                                                                                                                      'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Normality.__init__': ( 'src/probabilistic_methods.html#normality.__init__',
                                                                                                                               'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Normality.get_prediction_levels': ( 'src/probabilistic_methods.html#normality.get_prediction_levels',
                                                                                                                                            'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Normality.get_prediction_quantiles': ( 'src/probabilistic_methods.html#normality.get_prediction_quantiles',
                                                                                                                                               'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.Normality.get_samples': ( 'src/probabilistic_methods.html#normality.get_samples',
                                                                                                                                  'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU': ( 'src/probabilistic_methods.html#permbu',
                                                                                                                   'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU.__init__': ( 'src/probabilistic_methods.html#permbu.__init__',
                                                                                                                            'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU._nonzero_indexes_by_row': ( 'src/probabilistic_methods.html#permbu._nonzero_indexes_by_row',
                                                                                                                                           'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU._obtain_ranks': ( 'src/probabilistic_methods.html#permbu._obtain_ranks',
                                                                                                                                 'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU._permutate_predictions': ( 'src/probabilistic_methods.html#permbu._permutate_predictions',
                                                                                                                                          'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU._permutate_samples': ( 'src/probabilistic_methods.html#permbu._permutate_samples',
                                                                                                                                      'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU.get_prediction_levels': ( 'src/probabilistic_methods.html#permbu.get_prediction_levels',
                                                                                                                                         'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU.get_prediction_quantiles': ( 'src/probabilistic_methods.html#permbu.get_prediction_quantiles',
                                                                                                                                            'hierarchicalforecast/probabilistic_methods.py'),
                                                            'hierarchicalforecast.probabilistic_methods.PERMBU.get_samples': ( 'src/probabilistic_methods.html#permbu.get_samples',
                                                                                                                               'hierarchicalforecast/probabilistic_methods.py')},
            'hierarchicalforecast.utils': { 'hierarchicalforecast.utils.CodeTimer': ( 'src/utils.html#codetimer',
                                                                                      'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.CodeTimer.__enter__': ( 'src/utils.html#codetimer.__enter__',
                                                                                                'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.CodeTimer.__exit__': ( 'src/utils.html#codetimer.__exit__',
                                                                                               'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.CodeTimer.__init__': ( 'src/utils.html#codetimer.__init__',
                                                                                               'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.HierarchicalPlot': ( 'src/utils.html#hierarchicalplot',
                                                                                             'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.HierarchicalPlot.__init__': ( 'src/utils.html#hierarchicalplot.__init__',
                                                                                                      'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.HierarchicalPlot.plot_hierarchical_predictions_gap': ( 'src/utils.html#hierarchicalplot.plot_hierarchical_predictions_gap',
                                                                                                                               'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.HierarchicalPlot.plot_hierarchically_linked_series': ( 'src/utils.html#hierarchicalplot.plot_hierarchically_linked_series',
                                                                                                                               'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.HierarchicalPlot.plot_series': ( 'src/utils.html#hierarchicalplot.plot_series',
                                                                                                         'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.HierarchicalPlot.plot_summing_matrix': ( 'src/utils.html#hierarchicalplot.plot_summing_matrix',
                                                                                                                 'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils._construct_adjacency_matrix': ( 'src/utils.html#_construct_adjacency_matrix',
                                                                                                        'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils._is_strictly_hierarchical': ( 'src/utils.html#_is_strictly_hierarchical',
                                                                                                      'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils._lasso': ('src/utils.html#_lasso', 'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils._ma_cov': ( 'src/utils.html#_ma_cov',
                                                                                    'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils._shrunk_covariance_schaferstrimmer_no_nans': ( 'src/utils.html#_shrunk_covariance_schaferstrimmer_no_nans',
                                                                                                                       'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils._shrunk_covariance_schaferstrimmer_with_nans': ( 'src/utils.html#_shrunk_covariance_schaferstrimmer_with_nans',
                                                                                                                         'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils._to_upper_hierarchy': ( 'src/utils.html#_to_upper_hierarchy',
                                                                                                'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.aggregate': ( 'src/utils.html#aggregate',
                                                                                      'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.aggregate_temporal': ( 'src/utils.html#aggregate_temporal',
                                                                                               'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.get_cross_temporal_tags': ( 'src/utils.html#get_cross_temporal_tags',
                                                                                                    'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.is_strictly_hierarchical': ( 'src/utils.html#is_strictly_hierarchical',
                                                                                                     'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.level_to_outputs': ( 'src/utils.html#level_to_outputs',
                                                                                             'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.make_future_dataframe': ( 'src/utils.html#make_future_dataframe',
                                                                                                  'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.quantiles_to_outputs': ( 'src/utils.html#quantiles_to_outputs',
                                                                                                 'hierarchicalforecast/utils.py'),
                                            'hierarchicalforecast.utils.samples_to_quantiles_df': ( 'src/utils.html#samples_to_quantiles_df',
                                                                                                    'hierarchicalforecast/utils.py')}}}
