# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"HierarchicalReconciliation": "core.ipynb",
         "HierarchicalEvaluation": "evaluation.ipynb",
         "bottom_up": "methods.ipynb",
         "BottomUp": "methods.ipynb",
         "is_strictly_hierarchical": "methods.ipynb",
         "top_down": "methods.ipynb",
         "TopDown": "methods.ipynb",
         "middle_out": "methods.ipynb",
         "MiddleOut": "methods.ipynb",
         "crossprod": "methods.ipynb",
         "min_trace": "methods.ipynb",
         "MinTrace": "methods.ipynb",
         "optimal_combination": "methods.ipynb",
         "OptimalCombination": "methods.ipynb",
         "lasso": "methods.ipynb",
         "erm": "methods.ipynb",
         "ERM": "methods.ipynb",
         "aggregate": "utils.ipynb"}

modules = ["core.py",
           "evaluation.py",
           "methods.py",
           "utils.py"]

doc_url = "https://Nixtla.github.io/hierarchicalforecast/"

git_url = "https://github.com/Nixtla/hierarchicalforecast/tree/main/"

def custom_doc_links(name): return None
